// ROMAN ROBERT ALEXANDRU
// 333CB

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "nokia5110.h"
#include "nokia5110_chars.h"


/* definitii note (Hz) */
#define c3    131   // octava 3
#define C3    139
#define d3    147
#define D3    156
#define e3    165
#define f3    175
#define F3    185
#define g3    196
#define G3    208
#define a3    220
#define A3    233
#define b3    247
#define c4    261   // ocatava 4
#define C4    277
#define d4    294
#define D4    311
#define e4    329
#define f4    349
#define F4    370
#define g4    391
#define G4    415
#define a4    440
#define A4    466
#define b4    493
#define c5    523
#define C5    554
#define d5    587    // octava 5
#define D5    622
#define e5    659
#define f5    698
#define F5    740
#define g5    784
#define G5    830
#define a5    880
#define A5    932
#define b5    988

uint16_t dash = 100;
uint16_t surprise_notes[] = {d4, d4, a4, G4, g4, f4, d4, f4, g4, c4, c4, a4, G4, g4,
                             f4, d4, f4, g4, a4, G4, g4, f4, d4, f4, g4,
                             a4, G4, g4, f4, d4, f4, g4, d4, d4, a4, G4, g4, f4, d4,
                             f4, g4, c4, c4, a4, G4, g4, f4, d4, f4, g4, a4, G4,
                             g4, f4, d4, f4, g4, a4, G4, g4, f4, d4, f4, g4, d4, d4,
                             a4, G4, g4, f4, d4, f4, g4, c4, c4, a4, G4, g4, f4,
                             d4, f4, g4, a4, G4, g4, f4, d4, f4, g4, a4 
};              
uint16_t durations[] = {  dash, 3*dash, 3*dash, 2*dash, 2*dash, 2*dash,   dash,   dash,   dash,   dash, 3*dash, 3*dash, 2*dash, 2*dash,
                        2*dash,   dash,   dash, 5*dash, 3*dash, 2*dash, 2*dash, 2*dash,   dash,   dash, 5*dash,
                        3*dash, 2*dash, 2*dash, 2*dash,   dash,   dash,   dash,   dash, 3*dash, 3*dash, 2*dash, 2*dash, 2*dash,   dash,
                          dash,   dash,   dash, 3*dash, 3*dash, 2*dash, 2*dash, 2*dash,   dash,   dash, 5*dash, 3*dash, 2*dash,
                        2*dash, 2*dash,   dash,   dash, 5*dash, 3*dash, 2*dash, 2*dash, 2*dash,   dash,   dash,   dash,   dash, 3*dash,
                        3*dash, 2*dash, 2*dash, 2*dash,   dash,   dash,   dash,   dash, 3*dash, 3*dash, 2*dash, 2*dash, 2*dash,
                          dash,   dash, 5*dash, 3*dash, 2*dash, 2*dash, 2*dash,   dash,   dash, 5*dash, 3*dash 
};

uint8_t num_notes = sizeof(durations) / sizeof(durations[0]);


typedef struct Bullet {
    short x;
    short y;
    short exists;
} Bullet;

typedef struct Ship {
	short x;
	short y;
	short exists;
	short width;
	short height;
} Ship;


static int gamemode;
static Ship enemies[NO_SHIPS];
static Ship ship;
static int current_frame = 0;
static int nr_nota;
static int durata;
static int lives;
static Bullet bullets[NO_BULLETS];
static unsigned int rrandom = 0;

void draw_ship(Ship *ship) {
	for(int i = 0; i < ship->height; i++)
		for(int j = 0; j < ship->width; j++)
			if(pgm_read_byte(&(ship_shape[i][j]))) {
				nokia_lcd_set_pixel(j + ship->x, i + ship->y, 1);
			}
}

void IO_init()
{
	/* speaker-ul setat ca iesire */
	DDRD |= (1 << PD4);

	DDRC &= ~(1 << PC0);       // up
	DDRC &= ~(1 << PC1);       // down
	DDRC &= ~(1 << PC2);       // left
	DDRC &= ~(1 << PC3);       // right
	DDRC &= ~(1 << PC4);       // fire
	DDRC &= ~(1 << PC5);       // start | pause

	PORTC |= (1 << PC0);
	PORTC |= (1 << PC1);
	PORTC |= (1 << PC2);
	PORTC |= (1 << PC3);
    PORTC |= (1 << PC4);
    PORTC |= (1 << PC5);

	PCMSK2 |= (1 << PCINT16);
	PCMSK2 |= (1 << PCINT17);
	PCMSK2 |= (1 << PCINT18);
	PCMSK2 |= (1 << PCINT19);
    PCMSK2 |= (1 << PCINT20);
    PCMSK2 |= (1 << PCINT21);
}

void timer0_init()
{
	TCCR0A |= (1 << WGM01) | (1 << WGM00) | (1 << COM0A1) | (1 << COM0A0);
	OCR0A = 0xC0;
	TCCR0B |= (1 << CS00);
}

void timer1_init()
{

	TCCR1B |= (1 << CS10);
	TCCR1A |= (1 << COM1A1) | (1 << COM1A0);

	ICR1 = 0xFFFF;

	TCCR1A |= (1 << COM1B1) | (1 << WGM10) | (1 << WGM11);
	TCCR1B |= (1 << WGM12) | (1 << WGM13) | (1 << CS12);

	OCR1A = ICR1 / 4;
	OCR1B = ICR1 / 2;
}

void timer2_init()
{
	TCCR2A |= (1 << WGM21) | (1 << WGM20) | (1 << COM2A1) | (1 << COM2A0);
	TCCR2B |= (1 << CS20);
}

void update_song()
{
	durata -= 25;
	if (durata < 25) {
		OCR1A = 15625 / surprise_notes[nr_nota];
		OCR1B = OCR1A/2;
		TCNT1 = 0;
		durata = durations[nr_nota];
		nr_nota = (nr_nota+1)%num_notes;
	}
}



void draw_background(void) {
	int index;
	for(int i = 0; i < BACKGROUND_HEIGHT; i++)
		for(int j = 0; j < SCREEN_WIDTH; j++) {
			index = (j + current_frame/4) % BACKGROUND_WIDTH;
			if(pgm_read_byte(&(background[i][index]))) {
				nokia_lcd_set_pixel(j, SCREEN_HEIGHT-(BACKGROUND_HEIGHT-i), 1);
			}
		}
}

int check_collision_bullet(Bullet *bullet) {
	for (int enemy_index = 0; enemy_index < NO_SHIPS; enemy_index++) {
		if (enemies[enemy_index].exists && bullet->exists)
            if (enemies[enemy_index].x < bullet->x + 3 &&
    				enemies[enemy_index].x + enemies[enemy_index].width > bullet->x &&
    				enemies[enemy_index].y < bullet->y + 1 &&
    				enemies[enemy_index].height + enemies[enemy_index].y > bullet->y) {
    			enemies[enemy_index].exists = 0;
    			bullet->exists = 0;
                return 1;
        }
    }
	return 0;
}

int check_collision(Ship ship1, Ship ship2) {
	if (ship1.x < ship2.x + ship2.width &&
	   ship1.x + ship1.width > ship2.x &&
	   ship1.y < ship2.y + ship2.height &&
	   ship1.height + ship1.y > ship2.y) {
		return 1;
	}
	return 0;
}

void draw_lives(void) {
	for (int l = 0; l < lives; l++)
		for(int i = 0; i < LIFE_HEIGHT; i++)
			for(int j = 0; j < LIFE_WIDTH; j++)
				if(pgm_read_byte(&(life_shape[i][j]))) {
					nokia_lcd_set_pixel(j + 5 * l + 1, i, 1);
				}
}


void game_init(void) {
	ship.width = SHIP_WIDTH;
	ship.height = SHIP_HEIGHT;
	ship.x = 5;
	ship.y = 20;
	

	lives = 3;

	for (int i = 0; i < NO_BULLETS; i++) {
        bullets[i].exists = 0;
        bullets[i].x = SCREEN_WIDTH + 10;
    }
}

void create_bullet(void) {
    for (int i = 0; i < NO_BULLETS; i++) {
        if (bullets[i].exists != 1) {
            bullets[i].x = ship.x + 5;
            bullets[i].y = ship.y + 3;
            bullets[i].exists = 1;
            break;
        }
    }
}

void draw_bullets(void) {
    for (int bullet_index = 0; bullet_index < NO_BULLETS; bullet_index++) {
        if (bullets[bullet_index].exists != 1)
            continue;

        Bullet *bullet = &(bullets[bullet_index]);
		if (check_collision_bullet(bullet)) {
			bullet->exists = 0;
			bullet->x = SCREEN_WIDTH + 10;
			continue;
		}

        if (bullet->x > SCREEN_WIDTH - 5)
            bullet->exists = 0;
        bullet->x++;

        nokia_lcd_set_pixel(bullet->x  , bullet->y, 1);
		nokia_lcd_set_pixel(bullet->x+1, bullet->y, 1);
		nokia_lcd_set_pixel(bullet->x+2, bullet->y, 1);
    }
}

void init_enemy(Ship *enemy, short x, short y, short exists)
{
    enemy->width = SHIP_WIDTH;
    enemy->height = SHIP_HEIGHT;
    enemy->x = x;
    enemy->y = y;
    enemy->exists = exists;
}

void draw_enemies(void)
{
	for (int enemy_index = 0; enemy_index < NO_SHIPS; enemy_index++) {
		if (enemies[enemy_index].exists != 1)
			continue;
			
		if (check_collision(ship, enemies[enemy_index])) {
			enemies[enemy_index].exists = 0;
			lives--;
            enemies[enemy_index].exists = 0;
			continue;
		}
	   
		Ship *enemy = &(enemies[enemy_index]);

		if (enemy->x < 0) {
			enemy->exists = 0;
			continue;
		} else if (current_frame % 5 == 0)
			enemy->x -= 1;
			
		for(int i = 0; i < enemy->height; i++)
			 for(int j = 0; j < enemy->width; j++)
				if(pgm_read_byte(&(ship_shape[i][j]))) {
					nokia_lcd_set_pixel(enemy->width-1 - j + enemy->x, i + enemy->y, 1);
				}
	}
}


void create_enemies(void) {
    if (current_frame % 70 == 0) {
		rrandom++;
        for (int i = 0; i < NO_SHIPS; i++) {
            if (enemies[i].exists != 1) {
				srand(rrandom*current_frame);
                int x = 70;
                int y = rand() % (SCREEN_HEIGHT-10);
                init_enemy(&(enemies[i]), x, y, 1);
                break;
            }
        }
    }
}

void move_ship() {
    if (!(PINC & (1 << PC0)) && ship.y > 1) {
        ship.y--;
	}
    if (!(PINC & (1 << PC1)) && ship.y < 40) {
        ship.y++;
    }
    if (!(PINC & (1 << PC2)) && ship.y > 0) {
        ship.x--;
	}
    if (!(PINC & (1 << PC3)) && ship.y < 80) {
        ship.x++;
	}
    if (!(PINC & (1 << PC4))) {
        rrandom++;
        create_bullet();
    }
}

int main(void)
{
	nokia_lcd_init();
	sei();

	IO_init();
	timer1_init();
	timer2_init();

    gamemode = 0;
	OCR0A = 0;
	OCR1A = 0xFFFF;
	OCR2A = 0xFF;
	while (1) {
		rrandom++;
		nokia_lcd_clear();
        switch (gamemode) {
        // init display
        case 0:
                game_init();
                nokia_lcd_write_string("SPACE", 2);
                nokia_lcd_set_cursor(10, 15);
                nokia_lcd_write_string("IMPACT", 2);
                nokia_lcd_set_cursor(10, 30);
                nokia_lcd_write_string("Press START!", 1);
                if (!(PINC & (1 << PC5))) {
                    gamemode = 1;
                    _delay_ms(300);
                }
                break;
        // actual game
        case 1:
                if (!(PINC & (1 << PC5))) {
                    gamemode = 2;
                    _delay_ms(300);
                }
                draw_background();
                draw_lives();
                if (lives == 0) {
                    gamemode = 3;
                }
                draw_ship(&ship);
                create_enemies();
                draw_enemies();
                draw_bullets();
                move_ship();
                break;
        // pause
        case 2:
                nokia_lcd_set_cursor(10, 15);
                nokia_lcd_write_string("PAUSED", 2);
                if (!(PINC & (1 << PC5))) {
                    gamemode = 1;
                    _delay_ms(300);
                }
                break;
        // end screen
        case 3:
                nokia_lcd_set_cursor(10, 0);
                nokia_lcd_write_string("GAME", 2);
                nokia_lcd_set_cursor(25, 25);
                nokia_lcd_write_string("OVER", 2);
                if (!(PINC & (1 << PC5))) {
                    gamemode = 0;
                    _delay_ms(300);
                }
                break; 
        }
		nokia_lcd_render();
		update_song();
		if (rrandom > 1000000)
			rrandom = 0;
		current_frame++;
	}
}
